/* AMX Mod script.
*
*   FakeFull (fakefull_original.sma)
*   Copyright (C) 2003-2011  Freecode/JTP10181
*
*   This program is free software; you can redistribute it and/or
*   modify it under the terms of the GNU General Public License
*   as published by the Free Software Foundation; either version 2
*   of the License, or (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*   In addition, as a special exception, the author gives permission to
*   link the code of this program with the Half-Life Game Engine ("HL
*   Engine") and Modified Game Libraries ("MODs") developed by Valve,
*   L.L.C ("Valve"). You must obey the GNU General Public License in all
*   respects for all of the code used other than the HL Engine and MODs
*   from Valve. If you modify this file, you may extend this exception
*   to your version of the file, but you are not obligated to do so. If
*   you do not wish to do so, delete this exception statement from your
*   version.
*
****************************************************************************
*
*   Version 1.7.7        Date: 2011/01/17
*
*   Original Author: Freecode    freecode@hotmail.com
*   Current Author: JTP10181    jtp@jtpage.net
*
****************************************************************************
*
*  Based on OLO's FakeFull metamod plugin. Fake clients
*  connect to the server when its empty. Everytime someone
*  joins and there is no more spots left for another person
*  one fake client is removed. When the server is empty
*  and FakeFull is on Automatic mode it will add
*  a fake client. Once the # of fake clients and real players
*  equals to ff_players it will stop.
*  You need AMX Mod 2010.1 or higher.
*  VexdUM module must be enabled.
*
*  Commands:
*  amx_addfake <number|@FILL> - adds this many bots
*  amx_removefake <number|@ALL> - removes this many bots
*
*  Cvars: Paste the following into your amx.cfg to change defaults.
*  You must uncomment cvar lines for them to take effect
*
****************************************************************************
*  CVAR CONFIG BEGIN
****************************************************************************

// ******************  FakeFull Settings  ******************
// Turns Automatic mode on.
// ff_players must be set higher then 0.
// <1 = ON | 0 = OFF>
//ff_automode 0

// Minimal number of fake and real clients on the server.
// Fake clients will be kicked so that the total players on the
// Server equals this number.  If there are more players than this
// All the fake players will be gone. Do NOT set this at or aboive
// your max player count or no one will be able to join
// REQUIRED to be above 0 for Automatic mode.
//ff_players 2

// Delay between each fake client joins/leaves.
// REQUIRED to be above 0 for Automatic mode and regular bot add.
//ff_delay 1

****************************************************************************
*  CVAR CONFIG END
****************************************************************************
*
*  Setup for Automatic Mode:
*
*  Set ff_players to the number of fake clients you want in server at max.
*  Turn on ff_automode.
*
*  Changelog:
*
*  v1.7.7 - 11/01/17
*  - Ported to AMX Mod? Fixed number of bots added (one bot was too much at each usage of the cmds)
*
*  v1.7.6 - JTP10181 - 08/09/06
*  - Set team to UNASSIGNED for CS so they wont join teams
*
*  v1.7.5 - JTP10181 - 08/03/06
*  - Got rid of need for engine module
*  - Upgraded to pCVARS
*  - Finally put an end to fakes ending up with models and being seen as ghost players
*
*  v1.7.4 - JTP10181 - 07/01/06
*  - Fixed bug in team checking causing fakes on team "-1" to get kicked
*  - Upgraded to new file natives
*
*  v1.7.3 - JTP10181 - 4/21/06
*  - Added code to kick bots if they get on a team somehow
*
*  v1.7.2 - JTP10181
*  - Added new line of code per suggestion on a forum post, to prevent crashing
*
*  v1.7.1 - JTP10181
*  - Cleaned up massive code, hopefully no more crashing
*  - Renamed CVARS and commands
*  - Added feature to randomly add frags to the fake clients
*
*  v1.7 - JTP10181
*  - Ported fake client creation to new method using AMX. No longer needs module.
*  - Changed bot name file to botnames.cfg
*
*  v1.6.1 - JTP
*  - Fixed bug that could get server stuck in an infinite loop
*  - Fixed bug that caused dupe names sometimes
*
*  v1.6 - JTP
*  - Ported to AMXx
*  - Fixed mem leak/overflow bug caused by some of my code
*  - Fixed bugs where it was not removing the fakes automatically
*  - Rewrote a lot of the createBot function to make it more effecient
*    and less likely to cause mem overflows from looping
*  - Fixed bugs when setting the max_names higher.
*  - Must have the botnames.cfg now, no defaults are embedded in the script
*  - Fixed bug where playnames array was 1000 length for the names, made plugin huge.
*
*  v1.5
*  - Custom names and reading names from botnames.cfg
*
*  Below v1.6 was maintained by FreeCode
*
****************************************************************************/

#define FLAG_AMX_ADDFAKE    ADMIN_RCON
#define FLAG_AMX_REMOVEFAKE ADMIN_RCON

#include <translator>
#include <amxmod>
#include <amxmisc>
#include <fun>
#include <VexdUM>

new const Plugin[]  = "FakeFull Original"
new const Version[] = "1.7.7"
new const Author[]  = "JTP10181/Freecode/AssKicR"

#define MAX_NAMES   64
#define OFFSET_TEAM 114

new bIsUserFFBot, bIsUserConnected
new bool:checkingStatus = false
new bool:changingBots = false
new botCount = 0, namesread = 0
new namesToUse[MAX_NAMES][32]
new pDelay, pAuto, pPlayers
new g_iMaxPlayers
new g_bCstrikeRunning

public plugin_init() {
  load_translations("fakefull_original")
  register_plugin(_T(Plugin), Version, Author)
  pAuto = register_cvar("ff_automode", "0")
  pPlayers = register_cvar("ff_players", "2")
  pDelay = register_cvar("ff_delay", "1")
  register_concmd("amx_addfake", "botadd", FLAG_AMX_ADDFAKE, _T("<number|@FILL> - adds this many fake bots"))
  register_concmd("amx_removefake", "botremove", FLAG_AMX_REMOVEFAKE, _T("<number|@ALL> - removes this many fake bots"))
  set_task(10.0, "ServerStatus", 867254, _, _, "b")
  ReadNames()
  g_bCstrikeRunning = (is_running("cstrike") || is_running("czero"))
  g_iMaxPlayers = get_maxplayers()
}

public botadd(id, level, cid) {
  if(!cmd_access(id, level, cid, 1))
    return PLUGIN_HANDLED

  new arg[10], botNum
  if(read_argc() == 1) botNum = 1
  else {
    read_argv(1, arg, 9)
    if(equali(arg, "@FILL")) botNum = g_iMaxPlayers - get_playersnum(1)
    else botNum = str_to_num(arg)
  }
  if(botNum <= 0 || botNum > g_iMaxPlayers - get_playersnum(1)) {
    console_print(id, _T("[AMX Mod] Invalid number of bots to add."))
    return PLUGIN_HANDLED
  }
  new Float:botDelay = get_cvarptr_float(pDelay)
  console_print(id, _T("[AMX Mod] Adding %d bots with %.1f second delay for each one."), botNum, botDelay)
  set_task(botDelay, "createBot", 0, _, _, "a", botNum - 1)
  set_task(botDelay * (botNum + 1), "doneChanging")
  changingBots = true
  return PLUGIN_HANDLED
}

public botremove(id, level, cid) {
  if(!cmd_access(id, level, cid, 1))
    return PLUGIN_HANDLED

  new arg[10], botNum

  if(read_argc() == 1) botNum = 1
  else {
    read_argv(1, arg, 9)
    if(equali(arg, "@ALL")) botNum = botCount
    else botNum = str_to_num(arg)
  }
  if(botNum <= 0 || botNum > botCount) {
    console_print(id, _T("[AMX Mod] Invalid number of bots to remove."))
    return PLUGIN_HANDLED
  }
  new Float:botDelay = get_cvarptr_float(pDelay)
  console_print(id, _T("[AMX Mod] Removing %d bots with %.1f second delay for each one."), botNum, botDelay)
  set_task(botDelay, "removeBot", 0, _, _, "a", botNum - 1)
  set_task(botDelay * (botNum + 1), "doneChanging")
  changingBots = true
  return PLUGIN_HANDLED
}

public client_putinserver(id) {
  //Don't want anyone going invisible on us
  entity_set_int(id, EV_INT_rendermode, kRenderNormal)
  bIsUserFFBot &= ~(1<<(id&31))
  bIsUserConnected |= (1<<(id&31))
  ServerStatus()
}

public client_disconnect(id) {
  bIsUserFFBot &= ~(1<<(id&31))
  bIsUserConnected &= ~(1<<(id&31))
  ServerStatus()
}

public createBot() {
  new bool:UseAnotherName
  new BotName[32], UserNames[32][32]
  new name_rand = random_num(0, namesread - 1)
  new endLoop = name_rand - 1
  if(endLoop < 0) endLoop = namesread - 1

  //Save all the usernames so we dont have to keep getting them all
  for(new x = 1; x <= g_iMaxPlayers; x++) {
    if(!(bIsUserConnected & (1<<(x&31)))) continue
    get_user_name(x, UserNames[x], 31)
  }
  do {
    UseAnotherName = false
    copy(BotName, 31, namesToUse[name_rand])
    for(new id = 1; id <= g_iMaxPlayers; id++) {
      if(!(bIsUserConnected & (1<<(id&31)))) continue
      if(equali(BotName, UserNames[id])) {
        UseAnotherName = true
        if(name_rand == endLoop) {
          UseAnotherName = false
          log_amx(_T("ERROR: Ran out of names to use, please add more to the ^"botnames.cfg^"."))
        }
        name_rand++
        if(name_rand > namesread - 1) {
          name_rand = 0
        }
        break
      }
    }
  }
  while(UseAnotherName)

  new Bot = create_bot(BotName)
  if(Bot > 0) {
    //Supposed to prevent crashes?
    static const szGameEntity[] = "player"
    dllfunc(MetaFunc_CallGameEntity, szGameEntity, Bot)
    entity_set_int(Bot, EV_INT_flags, FL_FAKECLIENT)

    //Make Sure they have no model
    entity_set_string(Bot, EV_SZ_model, "")
    entity_set_string(Bot, EV_SZ_viewmodel, "")
    entity_set_int(Bot, EV_INT_modelindex, 0)

    //Make them invisible for good measure
    entity_set_int(Bot, EV_INT_renderfx, kRenderFxNone)
    entity_set_int(Bot, EV_INT_rendermode, kRenderTransAlpha)
    entity_set_float(Bot, EV_FL_renderamt, 0.0)

    //Set the team if we need to for this mod
    set_team(Bot)

    bIsUserFFBot |= (1<<(Bot&31))
    botCount++
  }
}

public removeBot() {
  for(new id = 1; id <= g_iMaxPlayers; id++) {
    if(bIsUserFFBot & (1<<(id&31))) {
      server_cmd("kick #%d", get_user_userid(id))
      bIsUserFFBot &= ~(1<<(id&31))
      botCount--
      return
    }
  }
}

public doneChanging() {
  changingBots = false
}

public ServerStatus() {
  if(!get_cvarptr_num(pAuto)) return
  if(checkingStatus || changingBots) return

  checkingStatus = true
  new rnd

  if(botCount > 0) {
    for(new id = 1; id <= g_iMaxPlayers; id++) {
      if(!(bIsUserConnected & (1<<(id&31)))) continue
      if(!(bIsUserFFBot & (1<<(id&31)))) continue
      rnd = random_num(1, 100)
      if(rnd <= 10) {
        set_user_frags(id, get_user_frags(id) + 1)
      }
      //Set the team if we need to for this mod
      set_team(id)
      if(get_user_team(id) > 0) {
        server_cmd("kick #%d", get_user_userid(id))
        bIsUserFFBot &= ~(1<<(id&31))
        botCount--
      }
    }
  }
  new pnum = get_playersnum(1)
  new ff_players = get_cvarptr_num(pPlayers)
  new Float:botDelay = get_cvarptr_float(pDelay)

  if(ff_players > g_iMaxPlayers - 2) {
    ff_players = g_iMaxPlayers - 2
    set_cvarptr_num(pPlayers, ff_players)
  }
  if(botDelay <= 0.0 ) {
    log_amx(_T("ERROR: Please set the cvar ^"ff_delay^" to a number higher than 0."))
  }
  else if(ff_players > pnum) {
    new addnum = ff_players - pnum
    set_task(botDelay, "createBot", 0, _, _, "a", addnum)
    set_task(botDelay * (addnum + 1), "doneChanging")
    changingBots = true
  }
  else if(ff_players < pnum) {
    new removenum = pnum - ff_players
    removenum = min(removenum, botCount)
    if(removenum > 0) {
      set_task(botDelay, "removeBot", 0, _, _, "a", removenum)
      set_task(botDelay * (removenum + 1), "doneChanging")
      changingBots = true
    }
  }
  checkingStatus = false
}

public set_team(BotID) {
  if(g_bCstrikeRunning) {
    set_offset_int(BotID, OFFSET_TEAM, false)
  }
}

ReadNames() {
  new szBotsNamesFile[64]
  build_path(szBotsNamesFile, charsmax(szBotsNamesFile), "$configdir/botnames.cfg")
  if(file_exists(szBotsNamesFile)) {
    new szText[32], iLine, iTextLen
    namesread = 0
    while(namesread < MAX_NAMES && read_file(szBotsNamesFile, iLine++, szText, charsmax(szText), iTextLen)) {
      if(!iTextLen || szText[0] == ';' || (szText[0] == '/' && szText[1] == '/') || szText[0] == '#') continue
      trim_imp(szText)
      namesToUse[namesread++] = szText
    }
  }
  else {
    log_amx(_T("ERROR: Unable to read file ^"%s^", it is required to load bot names."), szBotsNamesFile)
  }
}
